/*
 * geoLines.c - structures and procedures to manipulate
 *   geoLines (arrays of geoPoints) and geoLineArrays (arrays of geoLines)
 *
 * Copyright (C) 1999 Gordon D. Carrie
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 * 
 * Please address questions and bug reports to gcarri@mail.txcyber.com
 *
 ********************************************
 *
 */

#include <stdlib.h>
#include <float.h>
#include <math.h>
#include "geoProj.h"
#include "geoLines.h"

/* GeoLine - an array of geographical points */
struct GeoLineStrc
{
  unsigned	npts;		/* Number of points */
  unsigned	nptsMax;	/* Max number of points in current allocation */
  double	ordMax;		/* Northernmost ordinate for the line */
  double	ordMin;		/* Southernmost ordinate for the line */
  double	absMax;		/* Easternmost abscissa for the line */
  double	absMin;		/* Westernmost abscissa for the line */
  GeoPoint	*points;	/* Array of points in the line */
};

/* GeoLineArray - an array of GeoLines */
struct GeoLineArrayStrc
{
  unsigned	nLines;		/* Number of lines */
  unsigned	nLinesMax;	/* Max number of lines in current allocation */
  unsigned	npts;		/* Number of points for all lines */
  unsigned	nmax;		/* Number of points in longest line */
  double	ordMax;		/* Northernmost ordinate for all lines */
  double	ordMin;		/* Southernmost ordinate for all lines */
  double	absMax;		/* Easternmost abscissa for all lines */
  double	absMin;		/* Westernmost abscissa for all lines */
  struct GeoLineStrc *lines;	/* Pointer to the first line */
  unsigned	userCnt;	/* Number of widgets using this array */
};

/**************** Functions for GeoLines ****************/

void
InitGeoLine(struct GeoLineStrc *geoLine)
{
  /*
   *  Initialize an existing geoLine struct
   */
  
  geoLine->npts = geoLine->nptsMax = 0;
  geoLine->ordMax = geoLine->absMax = -FLT_MAX;
  geoLine->ordMin = geoLine->absMin = FLT_MAX;
  geoLine->points = NULL;
}

struct GeoLineStrc *
NewGeoLine(unsigned nptsMax)
{
  /*
   * Create a new geoLine with geoLine.nptsMax = nptsMax.
   * Returns NULL on failure.
   * Struct is dynamically allocated and must be freed by user.
   */

  struct GeoLineStrc *geoLine;			/* Return value */

  if ( !(geoLine = (struct GeoLineStrc *)malloc(sizeof(struct GeoLineStrc))) )
    return NULL;
  InitGeoLine(geoLine);
  if (nptsMax == 0)
    return geoLine;
  if ( !(geoLine->points = (GeoPoint *)malloc(nptsMax * sizeof(GeoPoint)))) {
    free(geoLine);		/* Allocation failed */
    return NULL;
  }
  geoLine->nptsMax = nptsMax;	/* Non-zero allocation requested and obtained */
  return geoLine;
}

void
ClearGeoLine(struct GeoLineStrc *geoLine)
{
  /*
   *  Initialize a geoLine without changing its memory allocation
   */

  geoLine->npts = 0;
  geoLine->ordMax = geoLine->absMax = -FLT_MAX;
  geoLine->ordMin = geoLine->absMin = FLT_MAX;
}

void
FreeGeoLine(struct GeoLineStrc *geoLine)
{
  /*
   *  Free storage associated with a geoLine
   */

   if (geoLine) free(geoLine->points);
   free(geoLine);
}

int
GrowGeoLine(struct GeoLineStrc *geoLine, unsigned nptsMax)
{
  /*
   * Set geoLine->nptsMax to nptsMax.
   * Returns 1 on success or 0 on failure.
   */
  
  GeoPoint *points;				/* Hold output from realloc */

  if (geoLine->nptsMax == nptsMax)
    return 1;
  points = (GeoPoint *)realloc(geoLine->points, nptsMax * sizeof(GeoPoint));
  if ( !points )
    return 0;			/* Allocation failed.  Leave alone */
  geoLine->nptsMax = nptsMax;	/* Allocation succeeded.  Update geoLine */
  geoLine->points = points;
  return 1;
}

int
AppendGeoPoint(GeoPoint geoPoint, struct GeoLineStrc *geoLine)
{
  /*
   * Add geoPoint to geoLine.
   * Returns 1 on success, 0 if (allocation) failure
   */

  unsigned incr = 100;	/* Allocation increment */

  if (   geoLine->npts + 1 > geoLine->nptsMax
      && !GrowGeoLine(geoLine, geoLine->nptsMax + incr))
    return 0;
  geoLine->ordMax = (geoLine->ordMax > geoPoint.ord)
                      ? geoLine->ordMax : geoPoint.ord;
  geoLine->ordMin = (geoLine->ordMin < geoPoint.ord)
                      ? geoLine->ordMin : geoPoint.ord;
  geoLine->absMax = (geoLine->absMax > geoPoint.abs)
                      ? geoLine->absMax : geoPoint.abs;
  geoLine->absMin = (geoLine->absMin < geoPoint.abs)
                      ? geoLine->absMin : geoPoint.abs;
  geoLine->points[geoLine->npts] = geoPoint;
  geoLine->npts++;
  return 1;
}

CartPoint GeoLineCtr(struct GeoLineStrc *geoLine)
{
  /* Return average of points in a geoline in Geocentric Cartesian coordinates*/

  CartPoint
    ctr = {0.0, 0.0, 0.0},	/* Hold return value */
    cpt;			/* Hold Cartesian coord's of one point */
  GeoPoint *gpt, *end;		/* Loop index */

  for (gpt = geoLine->points, end = gpt + geoLine->npts; gpt < end; gpt++) {
    cpt = LatLonToCart(*gpt);
    ctr.x += cpt.x;
    ctr.y += cpt.y;
    ctr.z += cpt.z;
  }
  ctr.x /= geoLine->npts;
  ctr.y /= geoLine->npts;
  ctr.z /= geoLine->npts;
  return ctr;
}

int GeoLineContainGeoPoint(GeoPoint geoPt, struct GeoLineStrc *geoln)
{
  /* Return 1 if geoPt in contained in region bounded by geoLine, else 0 */

  GeoPoint *pt0, *pt1, *end;	/* Loop indeces */
  int
    mrdx,	/* Number of times geoln crosses meridian containing geoPt */
    lnx;	/* Number of times geoln crosses line fm geoPt to North pole */
  enum LatSgn
    onMrd0, onMrd1;  /* PrMd if segment is on meridian containing geoPt */

  /* Loop through segments in line, counting number of times line
   * crosses meridian, and number of crossings between geoPt and North pole */
  for (mrdx = 0, lnx = 0,
         pt0 = geoln->points + geoln->npts - 1,
         onMrd0 = LonCmp(pt0->abs, geoPt.abs),
	 pt1 = geoln->points,
         end = geoln->points + geoln->npts;
       pt1 < end;
       pt0 = pt1++, onMrd0 = onMrd1) {

    /* Determine if segment defined by pt0--pt1 straddles meridian
     * containing geoPt, or is on boundary.  Do not count segments on
     * boundary as more than one crossing */
    if (LonBtwn1(geoPt.abs, pt1->abs, pt0->abs)) {

      double xlat;	/* Latitude of segment crossing */
      mrdx++;
      xlat = pt0->ord + (geoPt.abs - pt0->abs)
                        * (pt1->ord - pt0->ord) / (pt1->abs - pt0->abs);
      if (LatCmp(xlat, geoPt.ord) == North)
        lnx = !lnx;

    }
  }

  if (mrdx % 2 == 1) {

    /* Odd number of meridian crossings => region contains a pole */
    /* Assume pole is that of hemisphere containing line's mean */
    CartPoint ctr = GeoLineCtr(geoln);
    if (ctr.z > 0.0) {
      return !lnx;  /* Region contains North Pole */
    }

  }

  return lnx;
}

/* GeoLine member access functions */
unsigned GeoLineNPts(struct GeoLineStrc *geoLine) {return geoLine->npts;}

unsigned GeoLineNPtsMax(struct GeoLineStrc *geoLine) {return geoLine->nptsMax;}

double GeoLineOrdMax(struct GeoLineStrc *geoLine) {return geoLine->ordMax;}

double GeoLineOrdMin(struct GeoLineStrc *geoLine) {return geoLine->ordMin;}

double GeoLineAbsMax(struct GeoLineStrc *geoLine) {return geoLine->absMax;}

double GeoLineAbsMin(struct GeoLineStrc *geoLine) {return geoLine->absMin;}

GeoPoint
GetGeoPoint(struct GeoLineStrc *geoLine, unsigned n)
{
  /*
   *  Return GeoPoint n from geoLine, or a bogus location if out of bounds.
   */

   GeoPoint nowhere = NOWHERE;
   return (n < geoLine->npts) ? geoLine->points[n] : nowhere;
}

/**************** Functions for GeoLineArrays ****************/

struct GeoLineArrayStrc *
NewGeoLineArray(unsigned nLinesMax)
{
  /* 
   * Create a GeoLineArray with geoLineArray->nLinesMax = nLinesMax.
   * Returns NULL on failure.
   * Array is dynamically allocated and must be freed by user.
   */

  struct GeoLineArrayStrc *geoLineArray;	/* Return value */
  int n;					/* Loop index */

  geoLineArray
    = (struct GeoLineArrayStrc *)malloc(sizeof(struct GeoLineArrayStrc));
  if ( !geoLineArray )
    return NULL;
  geoLineArray->nLines = geoLineArray->nLinesMax = 0;
  geoLineArray->npts = geoLineArray->nmax = 0;
  geoLineArray->ordMax = geoLineArray->absMax = -FLT_MAX;
  geoLineArray->ordMin = geoLineArray->absMin = FLT_MAX;
  geoLineArray->lines = NULL;
  geoLineArray->userCnt = 0;
  if (nLinesMax == 0)
    return geoLineArray;
  geoLineArray->lines
    = (struct GeoLineStrc *)malloc(nLinesMax * sizeof(struct GeoLineStrc));
  if ( !geoLineArray->lines ) {
    free(geoLineArray);			/* malloc failed, no space */
    return NULL;
  }
  geoLineArray->nLinesMax = nLinesMax;	/* malloc succeeded */
  for (n = 0; n < nLinesMax; n++)
    InitGeoLine(geoLineArray->lines + n);
  return geoLineArray;
}

int
GrowGeoLineArray(struct GeoLineArrayStrc *lineArray, unsigned nLinesMax)
{
  /*
   * Set lineArray->nLinesMax = nLinesMax
   * Returns 1 on success or 0 on failure.
   */
  
  struct GeoLineStrc *lines;	/* Hold output from realloc */
  int n;			/* Loop index */

  if (lineArray->nLinesMax == nLinesMax)
    return 1;
  lines = (struct GeoLineStrc *)
            realloc(lineArray->lines, nLinesMax * sizeof(struct GeoLineStrc));
  if ( !lines )
    return 0;
  lineArray->nLinesMax = nLinesMax;
  lineArray->lines = lines;
  for (n = lineArray->nLines; n < lineArray->nLinesMax; n++)
    InitGeoLine(lineArray->lines + n);
  return 1;
}

void
GeoLineArrayAddUser(struct GeoLineArrayStrc *geoLineArray)
{
  /* Increment usercnt in geoLineArray.  */
  geoLineArray->userCnt++;
}

void
GeoLineArrayDropUser(struct GeoLineArrayStrc *geoLineArray)
{
  /* Decrement usercnt in geoLineArray */
  geoLineArray->userCnt--;
}

int
AppendGeoLine(struct GeoLineStrc *geoLine,struct GeoLineArrayStrc *geoLineArray)
{
  /*
   * Copy geoLine onto the end of geoLineArray
   */

  unsigned incr = 100;	/* Allocation increment */
  int np;		/* Loop index */
  int nLines;		/* Initial number of lines in geoLineArray */

  nLines = geoLineArray->nLines;
  if (   nLines + 1 > geoLineArray->nLinesMax
      && !GrowGeoLineArray(geoLineArray, geoLineArray->nLinesMax + incr))
    return 0;
  geoLineArray->npts += geoLine->npts;
  geoLineArray->nmax = (geoLineArray->nmax > geoLine->npts)
                         ? geoLineArray->nmax : geoLine->npts;
  geoLineArray->ordMax = (geoLineArray->ordMax > geoLine->ordMax)
                           ? geoLineArray->ordMax : geoLine->ordMax;
  geoLineArray->ordMin = (geoLineArray->ordMin < geoLine->ordMin)
                           ? geoLineArray->ordMin : geoLine->ordMin;
  geoLineArray->absMax = (geoLineArray->absMax > geoLine->absMax)
                           ? geoLineArray->absMax : geoLine->absMax;
  geoLineArray->absMin = (geoLineArray->absMin < geoLine->absMin)
                           ? geoLineArray->absMin : geoLine->absMin;
  GrowGeoLine(geoLineArray->lines + nLines, geoLine->npts);
  for (np = 0; np < geoLine->npts; np++) {
    geoLineArray->lines[nLines].points[np].ord = geoLine->points[np].ord;
    geoLineArray->lines[nLines].points[np].abs = geoLine->points[np].abs;
  }
  geoLineArray->lines[nLines].npts = geoLine->npts;
  geoLineArray->lines[nLines].ordMax = geoLine->ordMax;
  geoLineArray->lines[nLines].ordMin = geoLine->ordMin;
  geoLineArray->lines[nLines].absMax = geoLine->absMax;
  geoLineArray->lines[nLines].absMin = geoLine->absMin;
  geoLineArray->nLines++;
  return 1;
}

int
GeoLineArrayContainGeoPoint(GeoPoint geoPt,
                            struct GeoLineArrayStrc *geoLnArr)
{
  /* Return true if any geoline in geoLnArr contains geoPt */
  struct GeoLineStrc *geoLn, *end;		/* Loop pointers */

  for (geoLn = geoLnArr->lines, end = geoLnArr->lines + geoLnArr->nLines;
       geoLn < end;
       geoLn++) {
    if ( GeoLineContainGeoPoint(geoPt, geoLn) )
      return 1;
  }
  return 0;
}

/* Member access */
unsigned
GeoLineArrayNLines(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->nLines;
}

unsigned
GeoLineArrayNLinesMax(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->nLinesMax;
}

unsigned
GeoLineArrayNPts(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->npts;
}

unsigned
GeoLineArrayNmax(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->nmax;
}

double
GeoLineArrayOrdMax(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->ordMax;
}

double
GeoLineArrayOrdMin(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->ordMin;
}

double
GeoLineArrayAbsMax(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->absMax;
}

double
GeoLineArrayAbsMin(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->absMin;
}

unsigned
GeoLineArrayUserCnt(struct GeoLineArrayStrc *geoLineArray) {
  return geoLineArray->userCnt;
}

struct GeoLineStrc *
GetGeoLine(struct GeoLineArrayStrc *geoLineArray, unsigned n)
{
  /*
   *  Return GeoLine index n from geoLineArray, or NULL if out of bounds.
   */
  return (n < geoLineArray->nLines)
            ? geoLineArray->lines + n : (struct GeoLineStrc *)NULL;
}

void
FreeGeoLineArray(struct GeoLineArrayStrc *lineArray)
{
  /*
   * geoFreeLines - free storage for a lineArray 
   */
  
  int n;

  if ( !lineArray ) return;
  for (n = 0; n < lineArray->nLines; n++)
    free(lineArray->lines[n].points);
  free(lineArray->lines);
  free(lineArray);
  return;
}
