/*
 * geoProj.c - cartographic projections
 *
 * Copyright (C) 1999 Gordon D. Carrie
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 * 
 * Please address questions and bug reports to gcarri@mail.txcyber.com
 *
 ********************************************
 *
 */

/*
 * The functions in this file convert back and forth between latitude-longitude
 * and projection coordinates.  Projections are defined by various
 * parameters, which vary in both type and number according to projection
 * properties and geometry.  They also use different functions to make
 * conversions.  Each projection stores this information in a Projection
 * structure.  The variable parameters are in a union in this structure called
 * "prm."  The Projection structure also contains parameters common to all
 * projections, such as a character descriptor, scale, rotation angle, etc.,
 * as well as projection-specific functions for making conversions.  The header
 * file declares a Projection pointer, which points to a Projection structure.
 * This allows external functions to store and use Projection structures,
 * but only local functions have knowledge of and access to the structure
 * members.  Global functions such as LatLonToProj and ProjToLatLon take a
 * Projection pointer as an argument, extract the specific conversion function,
 * which they then call to make the actual conversion.  The specific conversion
 * function also takes a Projection pointer as an argument, providing the
 * function with the specific parameters needed to make the convresion.  The
 * Projection pointer is an "opaque pointer."  Although using it means extra
 * dereferencing and function calling, it makes it possible for a small number
 * of global functions to handle all kinds of projections, regardless of the
 * number and type of parameters.  This is one way of interfacing and data
 * hiding in C (although it might be worth redoing this part in C++ some time).
 *
 * To add a projection to this package, perform the following steps:
 * (Hint: before taking each step, examine the structures and functions in
 * this file in case what you is already here.)
 *   Declare in this file a struct with the parameters specifically needed by
 *     your projection (if not already here).
 *   Add this struct to the prm union in the Projection struct.
 *   Define a local function of form:
 *     GeoPoint (*latLonToProj)(GeoPoint, struct Projection*)
 *     to convert a lat-lon GeoPoint to a projection Geopoint.
 *     The global function, LatLonToProj, defined below, will use this function
 *     to convert lat-lons.
 *   Define a local function of form:
 *     GeoPoint (*projToLatLon)(GeoPoint, struct Projection*)
 *     to convert a projection GeoPoint to a lat-lon GeoPoint.
 *     The global function, ProjToLatLon, defined below, will use this function
 *     to convert projection points.
 *   Define a local function of form:
 *     GeoLineArray (*latLonArrayToProj)(GeoLineArray, struct Projection*)
 *     to convert an array of lat-lon points to projection points, taking into
 *     account the geometry of the projection domain (so some segments will
 *     be split or omitted).
 *     The global function, LatLonArrayToProj, defined below, will use this
 *     function to convert arrays of lat-lons.  Refer to GeoLines.c and 
 *     GeoLines.h for information about GeoLineArrays.
 *   Define a global function, whose argument list will vary, that sets
 *     the values of a projection struct to those appropriate for your
 *     projection.  It will need to set descriptor, prm, latLonToProj,
 *     projToLatLon, and latLonArrayToProj.
 *   Add this function to geoProj.h so users can create instances of
 *     your projection in external functions.
 *   If it works (or doesn't and you need help) and you are so inclined,
 *     email gcarri@mail.txcyber.com with your projection for inclusion in
 *     future versions.
 *
 */

/*
 *  Formulas and notation from:
 *    Snyder, John P.
 *    Map Projections used by the U.S. Geological Survey.
 *    (Geological Survey bulletin ; 1532)
 *    United States Government Printing Office, Washington:  1982.
 */

#include <math.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "geography.h"
#include "geoProj.h"
#include "geoLines.h"

#define DESCRLEN 200

/* Projection specific structures to use in union prm of struct Projection */

/* Reference point */
struct refPt {
  GeoPoint pt;		/* Reference point */
  double cosRLat;	/* Cosine of reference latitude */
  double sinRLat;	/* Sine of reference latitude */
};

/* Lambert conformal conic parameters */
struct lcc {
  double rlat;		/* reference latitude */
  double rlon;		/* vertical meridian */
  double n;		/* }                             */
  double RF;		/* Constants from Snyder, p. 105 */
  double rho0;		/* }                             */
};

/********* End of projection specific structures *********/

/* General projection information and parameters */
struct Projection
{
  /* Parameters common to all projections */
  char *descriptor;		/* Text info about projection */
  double scale;			/* Projection coordinates to hardware */
  double rotation;		/* Rotation angle (degrees cw from N) */
  double scosr;			/* scale * cos(rotation) */
  double ssinr;			/* scale * sin(rotation) */
  /* Projection specific parameters */
  union {
    enum LatSgn  ns;		/* Northern or southern hemisphere */
    double       rlon;		/* Reference longitude */
    struct refPt refPt;		/* Reference point */
    struct lcc   lcc;		/* Lambert Conformal Conic */
    } prm;
  /* Convert points */
  GeoPoint (*latLonToProj)(GeoPoint, struct Projection*);
  GeoPoint (*projToLatLon)(GeoPoint, struct Projection*);
  /* Convert array of lat-lon lines */
  GeoLineArray (*latLonArrayToProj)(GeoLineArray, struct Projection*);
};

/* Projection specific functions */

/* Cylindrical Equidistant */

static GeoPoint
latLonToCylEqdist(GeoPoint latLon, struct Projection * proj);

static GeoPoint
cylEqdistToLatLon(GeoPoint projPoint, struct Projection * proj);

/* Cylindrical Equal Area */

static GeoPoint
latLonToCylEqArea(GeoPoint latLon, struct Projection * proj);

static GeoPoint
cylEqAreaToLatLon(GeoPoint projPoint, struct Projection * proj);

/* Mercator */

static GeoPoint
latLonToMercator(GeoPoint latLon, struct Projection * proj);

static GeoPoint
mercatorToLatLon(GeoPoint projPoint, struct Projection * proj);

/* Lambert Conformal Conic */

static GeoPoint
latLonToLambertConfConic(GeoPoint latLon, struct Projection * proj);

static GeoPoint
lambrtConfConicToLatLon(GeoPoint projPoint, struct Projection * proj);

static GeoLineArray
latLonArrayToLambertConfCon(GeoLineArray latLonArray, struct Projection * proj);

/* Polar Stereographic */

static GeoPoint
latLonToPolarStereographic(GeoPoint latLon, struct Projection * proj);

static GeoPoint
polarStereographicToLatLon(GeoPoint projPoint, struct Projection * proj);

static GeoLineArray
latLonArrayToPolarSter(GeoLineArray latLonArray, struct Projection * proj);

/* Sinusoidal */

static GeoPoint
latLonToSinusoidal(GeoPoint latLon, struct Projection * proj);

static GeoPoint
sinusoidalToLatLon(GeoPoint projPoint, struct Projection * proj);

/* Orthographic */

static GeoPoint
latLonToOrthographic(GeoPoint latLon, struct Projection * proj);

static GeoPoint
orthographicToLatLon(GeoPoint projPoint, struct Projection * proj);

/* Multi-use */

static GeoLineArray
latLonArrayToCylProj(GeoLineArray latLonArray, struct Projection * proj);

static GeoLineArray
latLonArrayToHemisphere(GeoLineArray latLonArray, struct Projection * proj);

/************* End of declarations ************/

struct Projection *
AllocProj()
{
  /* Allocate storage for a projection.  Return null on failure. */

  struct Projection *proj;		/* Hold return value */
  
  if (   !(proj = (struct Projection*)malloc(sizeof(struct Projection)))
      || !(proj->descriptor = (char *)malloc(DESCRLEN)) ) {
    free(proj);
    proj = NULL;
  } else {
    /* Set to a default */
    GeoProjSetScale(proj, 1.0e-7);
    GeoProjSetRotation(proj, 0.0);
    SetCylEqdist(proj, 0.0);
  }
  return proj;
}

void
FreeProj(Projection proj)
{
  /* Free storage associated with a projection */
  if( proj ) {
    free(proj->descriptor);
  }
  free(proj);
}

GeoPoint
ProjToLatLon(GeoPoint projPoint, Projection proj)
{
  /*
   * Return lat-lon coordinates corresponding to projection coordinates
   * given by projPoint for projection given by proj
   */

  double
    cosr, sinr,		/* Cosine and sine */
    abs, ord;		/* Hold results of calculations */

  cosr = cos(RADPERDEG * proj->rotation);
  sinr = sin(RADPERDEG * proj->rotation);
  ord = (projPoint.abs * sinr + projPoint.ord * cosr) / proj->scale;
  abs = (projPoint.abs * cosr - projPoint.ord * sinr) / proj->scale;
  projPoint.ord = ord;
  projPoint.abs = abs;
  return proj->projToLatLon(projPoint, proj);
}

GeoPoint
LatLonToProj(GeoPoint latLon, Projection proj)
{
  /*
   * latLonToProj - return projection coordinates corresponding to
   *   lat-lon coordinates given by latLon for projection given by proj
   */

  double ord, abs;		/* Hold results of calculations */
  GeoPoint projPoint;		/* Hold output value */
 
  projPoint = proj->latLonToProj(latLon, proj);
  ord = projPoint.ord * proj->scosr - projPoint.abs * proj->ssinr;
  abs = projPoint.abs * proj->scosr + projPoint.ord * proj->ssinr;
  projPoint.ord = ord;
  projPoint.abs = abs;
  return projPoint;
}

GeoLineArray
LatLonArrayToProj(GeoLineArray latLonArray, Projection proj)
{
  /* Convert an array of latitude-longitude lines to projection coordinates */

  return proj->latLonArrayToProj(latLonArray, proj);
}

double
GeoProjGetScale(Projection proj)
{
  /* Return projection scale */
  return proj->scale;
}

void
GeoProjSetScale(Projection proj, double scale)
{
  /* Set scale in a projection */
  proj->scale = scale;
  proj->scosr = scale * cos(RADPERDEG * proj->rotation);
  proj->ssinr = scale * sin(RADPERDEG * proj->rotation);
}

double
GeoProjGetRotation(Projection proj)
{
  /* Return projection rotation */
  return proj->rotation;
}

void
GeoProjSetRotation(Projection proj, double rotation)
{
  /* Set rotation in a projection */
  proj->rotation = rotation;
  proj->scosr = proj->scale * cos(RADPERDEG * rotation);
  proj->ssinr = proj->scale * sin(RADPERDEG * rotation);
}

char *
ProjDescriptor(Projection proj)
{
  return proj->descriptor;
}

/************ Cylindrical Equidistant projection ************/

void
SetCylEqdist(struct Projection *proj, double rlon)
{
  /* Set proj to cylindrical equidistant centered at rlon */
  proj->prm.rlon = GwchLon(rlon);
  proj->latLonToProj = latLonToCylEqdist;
  proj->projToLatLon = cylEqdistToLatLon;
  proj->latLonArrayToProj = latLonArrayToCylProj;
  sprintf(proj->descriptor, "\"Cylindrical equidistant\" "
    "centered at %.3f deg lon", proj->prm.rlon);
}

GeoPoint
latLonToCylEqdist(GeoPoint latLon, struct Projection * proj)
{
  GeoPoint projPoint;

  projPoint.ord = CMPERDEG * latLon.ord;
  projPoint.abs = CMPERDEG * DomainLon(latLon.abs, proj->prm.rlon);
  return projPoint;
}

GeoPoint
cylEqdistToLatLon(GeoPoint projPoint, struct Projection * proj)
{
  GeoPoint latLon;

  latLon.ord = DEGPERCM * projPoint.ord;
  latLon.abs = DomainLon(DEGPERCM * projPoint.abs, proj->prm.rlon);
  return latLon;
}

/************ Mercator projection ************/

void
SetMercator(struct Projection *proj, double rlon)
{
  /* Set proj to Mercator projection */
  proj->prm.rlon = GwchLon(rlon);
  proj->latLonToProj = latLonToMercator;
  proj->projToLatLon = mercatorToLatLon;
  proj->latLonArrayToProj = latLonArrayToCylProj;
  sprintf(proj->descriptor, "Mercator centered at %.3f deg lon",
          proj->prm.rlon);
}

GeoPoint
latLonToMercator(GeoPoint latLon, struct Projection * proj)
{
  GeoPoint projPoint;
  
  latLon.ord = (latLon.ord == 90.0) ? 89.0
                                    : (latLon.ord == -90.0) ? -89.9
				                            : latLon.ord;
  projPoint.ord = REARTH * log(tan(M_PI_4 + 0.5 * RADPERDEG * latLon.ord));
  projPoint.abs = CMPERDEG * DomainLon(latLon.abs, proj->prm.rlon);
  return projPoint;
}

GeoPoint
mercatorToLatLon(GeoPoint projPoint, struct Projection * proj)
{
  GeoPoint latLon;

  latLon.ord = 2.0 * DEGPERRAD
               * (atan(exp(projPoint.ord * REARTH_INV)) - M_PI_4);
  latLon.abs = DomainLon(projPoint.abs * DEGPERCM, proj->prm.rlon);
  return latLon;
}

/************ Cylindrical Equal Area projection ************/

void
SetCylEqArea(struct Projection *proj, double rlon)
{
  /* Set proj to cylindrical equal area centered on rlon */
  proj->prm.rlon = GwchLon(rlon);
  proj->latLonToProj = latLonToCylEqArea;
  proj->projToLatLon = cylEqAreaToLatLon;
  proj->latLonArrayToProj = latLonArrayToCylProj;
  sprintf(proj->descriptor, "\"Cylindrical equal area\" "
    "centered at %.3f deg lon", proj->prm.rlon);
}

GeoPoint
latLonToCylEqArea(GeoPoint latLon, struct Projection * proj)
{
  GeoPoint projPoint;
  
  projPoint.ord = REARTH * sin(latLon.ord * RADPERDEG);
  projPoint.abs = CMPERDEG * DomainLon(latLon.abs, proj->prm.rlon);
  return projPoint;
}

GeoPoint
cylEqAreaToLatLon(GeoPoint projPoint, struct Projection * proj)
{
  GeoPoint latLon;

  latLon.ord = DEGPERRAD * asin(REARTH_INV * projPoint.ord);
  latLon.abs = DomainLon(DEGPERCM * projPoint.abs, proj->prm.rlon);
  return latLon;
}

/************ Lambert Conformal Conic projection ************/

void
SetLambertConfConic(struct Projection *proj, double rlat, double rlon)
{
  /* Set proj to Lambert Conformal Conic referenced to (rlat, rlon).  */
  /* Notation from Snyder, p. 105 */

  double
    phi0 = rlat * RADPERDEG,
    n = sin(phi0);
  double
    tan_n = pow(tan(M_PI_4 + 0.5 * phi0), n);
  double
    RF = REARTH * cos(phi0) * tan_n / n;

  proj->prm.lcc.rlat = rlat;
  proj->prm.lcc.rlon = GwchLon(rlon);
  proj->prm.lcc.n = n;
  proj->prm.lcc.RF = RF;
  proj->prm.lcc.rho0 = REARTH / tan(phi0);
  proj->projToLatLon = lambrtConfConicToLatLon;
  proj->latLonToProj = latLonToLambertConfConic;
  proj->latLonArrayToProj = latLonArrayToLambertConfCon;
  sprintf(proj->descriptor, "\"Lambert Conformal Conic\" "
    "reference lat = %.3f reference lon = %.3f", rlat, proj->prm.lcc.rlon);
}

GeoPoint
latLonToLambertConfConic(GeoPoint latLon, struct Projection * proj)
{
  /* Convert lat-lon to Lambert Conformal Conic */
  /* Formulas from Snyder, p. 105 */

  GeoPoint projPoint;  
  double dLon, theta, rho;

  /* Point at a pole */
  if (latLon.ord > 89.0 || latLon.ord < -89.0) {
    projPoint.ord = proj->prm.lcc.rho0;
    projPoint.abs = 0.0;
    return projPoint;
  }

  /* Avoid opposite pole */
  latLon.ord = (proj->prm.lcc.rlat > 0.0 && latLon.ord < -80.0)
                ? -80.0 : (proj->prm.lcc.rlat < 0.0 && latLon.ord > 80.0)
		           ? 80.0 : latLon.ord;

  dLon = DomainLon(latLon.abs, proj->prm.lcc.rlon) - proj->prm.lcc.rlon;
  theta = proj->prm.lcc.n * dLon * RADPERDEG;
  rho = proj->prm.lcc.RF / pow(tan(M_PI_4 + 0.5 * RADPERDEG *latLon.ord),
                               proj->prm.lcc.n);
  projPoint.abs = rho * sin(theta);
  projPoint.ord = proj->prm.lcc.rho0 - rho * cos(theta);
  return projPoint;
}

GeoPoint
lambrtConfConicToLatLon(GeoPoint projPt, struct Projection * proj)
{
  /* Convert Lambert Conformal Conic to lat-lon */
  /* Formulas from Snyder, p. 105 */

  GeoPoint latLon;
  double
     RF = proj->prm.lcc.RF,
      n = proj->prm.lcc.n;
  double
      x = n > 0.0 ? projPt.abs : -projPt.abs,
     dy = n > 0.0 ? proj->prm.lcc.rho0 - projPt.ord
                  : projPt.ord - proj->prm.lcc.rho0;
  double
    rho = n > 0.0 ? sqrt(projPt.abs * projPt.abs + dy * dy)
                  : -sqrt(projPt.abs * projPt.abs + dy * dy);
  double theta = atan2(x, dy) * DEGPERRAD;

  latLon.ord = (2.0 * atan(pow(RF/rho, 1.0/n)) - M_PI_2) * DEGPERRAD;
  latLon.abs = DomainLon(proj->prm.lcc.rlon + theta / n, proj->prm.lcc.rlon);
  return latLon;
}

static GeoLineArray
latLonArrayToLambertConfCon(GeoLineArray latLonArray, struct Projection * proj)
{
  /*
   * Put array of lat-lon's into Lambert Conformal Conic coordinates.
   * Essentially a copy of latLonArrayToCylProj.
   *
   * Later need to modify so projection is only in one hemisphere
   * (by convention).  gc 1999/06/18
   *
   * Returns a new geoLineArray with latLons in projection coordinates.
   * Storage for this array is dynamically allocated and must be freed by the
   * user.  Returns NULL on failure.
   */

  GeoLineArray projLineArray = NULL;	/* Return value */
  int nl, nls, np;			/* Loop parameters */
  double
    rlon = proj->prm.lcc.rlon,		/* Reference longitude */
    west = rlon - 180.0,
    east = rlon + 180.0,		/* West and east ends of domain */
    lon0 = 0.0,
    lon1 = 0.0;				/* Longitudes of last and current pt */
  GeoPoint
    latLon0, latLon1,			/* Points in the current segment */
    cross;				/* Point on domain boundary */
  GeoLine projLine = NULL;		/* Line being made right now */

  nls = GeoLineArrayNLines(latLonArray);
  if (   !(projLineArray = NewGeoLineArray(nls))
      || !(projLine = NewGeoLine(100)))
    goto error;

  for (nl = 0; nl < nls; nl++) {
    /* Loop through the lines.  Convert each line to a projection coordinates */
    GeoLine latLons = GetGeoLine(latLonArray, nl);

    latLon0 = GetGeoPoint(latLons, 0);
    AppendGeoPoint(LatLonToProj(latLon0, proj), projLine);
    for (np = 1; np < GeoLineNPts(latLons); np++) {
      latLon1 = GetGeoPoint(latLons, np);

      /* Do bookkeeping for segments crossing or touching boundaries */

      if (rlon > 0) {

        /* Check for segments crossing or touching western boundary */

	lon0 = DomainLon(latLon0.abs, west);
	lon1 = DomainLon(latLon1.abs, west);

	if (LonBtwn(west, lon0, lon1)) {

  	  /* Segment crosses longitude discontinuity */
  	  cross.ord = latLon0.ord + (latLon1.ord - latLon0.ord)
  	                            / (lon1 - lon0) * (west - lon0);
  	  if (LonCmp(lon0, west) == West) {

  	    /* Segment is crossing singularity from west to east */
  	    cross.abs = east;
            AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            if (GeoLineNPts(projLine) > 1)
              AppendGeoLine(projLine, projLineArray);
  	    ClearGeoLine(projLine);
  	    cross.abs = west;
            AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

  	  } else {

  	    /* Segment is crossing singularity from east to west */
  	    cross.abs = west;
            AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            if (GeoLineNPts(projLine) > 1)
              AppendGeoLine(projLine, projLineArray);
  	    ClearGeoLine(projLine);
  	    cross.abs = east;
            AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);
  	  }

        } else if (LonCmp(lon0, west) == West && LonCmp(lon1, west) == PrMd) {

  	  /* Segment is terminating at eastern boundary */
	  cross.ord = latLon1.ord;
  	  cross.abs = east;
  	  AppendGeoPoint(LatLonToProj(cross, proj), projLine);
          if (GeoLineNPts(projLine) > 1)
            AppendGeoLine(projLine, projLineArray);
          ClearGeoLine(projLine);
  	  AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        } else if (LonCmp(lon0, west) == PrMd && LonCmp(lon1, west) == West) {

          /* Segment is starting at eastern boundary and going west */
          if (GeoLineNPts(projLine) > 1)
            AppendGeoLine(projLine, projLineArray);
          ClearGeoLine(projLine);
          latLon0.abs = east;
          AppendGeoPoint(LatLonToProj(latLon0, proj), projLine);
          AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        } else {

	  if (DegCmp(lon0, east) == 0 && DegCmp(lon1, west) == 0)
	    latLon1.abs = east;
	  if (DegCmp(lon0, west) == 0 && DegCmp(lon1, east) == 0)
	    latLon1.abs = west;
          AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        }

      } else {	/* rlon <= 0 */

  	lon0 = DomainLon(latLon0.abs, east);
  	lon1 = DomainLon(latLon1.abs, east);

	if (LonBtwn(east, lon0, lon1)) {

  	  /* Segment crosses longitude discontinuity */
  	  cross.ord = latLon0.ord + (latLon1.ord - latLon0.ord)
  	                            / (lon1 - lon0) * (east - lon0);
	  if (LonCmp(lon0, east) == West) {
	    /* Segment is crossing singularity from west to east */
	    cross.abs = east;
	    AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            if (GeoLineNPts(projLine) > 1)
              AppendGeoLine(projLine, projLineArray);
	    ClearGeoLine(projLine);
	    cross.abs = west;
	    AppendGeoPoint(LatLonToProj(cross, proj), projLine);
	    AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);
          } else {
	    /* Segment is crossing singularity from east to west */
	    cross.abs = west;
	    AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            if (GeoLineNPts(projLine) > 1)
              AppendGeoLine(projLine, projLineArray);
	    ClearGeoLine(projLine);
	    cross.abs = east;
	    AppendGeoPoint(LatLonToProj(cross, proj), projLine);
	    AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);
          }

        } else if (LonCmp(lon0, east) == PrMd && LonCmp(lon1, east) == East) {

	  /* Segment is starting at western boundary and going east */
	  if (GeoLineNPts(projLine) > 1)
	    AppendGeoLine(projLine, projLineArray);
	  ClearGeoLine(projLine);
	  latLon0.abs = west;
	  AppendGeoPoint(LatLonToProj(latLon0, proj), projLine);
	  AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        } else if (LonCmp(lon0, east) == East && LonCmp(lon1, east) == PrMd) {

	  /* Segment is terminating at western boundary */
	  cross.ord = latLon1.ord;
  	  cross.abs = west;
  	  AppendGeoPoint(LatLonToProj(cross, proj), projLine);
          if (GeoLineNPts(projLine) > 1)
            AppendGeoLine(projLine, projLineArray);
          ClearGeoLine(projLine);
  	  AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        } else {

	  if (DegCmp(lon0, east) == 0 && DegCmp(lon1, west) == 0)
	    latLon1.abs = east;
	  if (DegCmp(lon0, west) == 0 && DegCmp(lon1, east) == 0)
	    latLon1.abs = west;
          AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        }
      }
      latLon0 = latLon1;
    }
    if (GeoLineNPts(projLine) > 1)
      AppendGeoLine(projLine, projLineArray);
    ClearGeoLine(projLine);
  }

  FreeGeoLine(projLine);
  return projLineArray;

  error:
  FreeGeoLine(projLine);
  FreeGeoLineArray(projLineArray);
  return NULL;
}

/************ Polar Stereographic projection ************/

void
SetPolarStereographic(struct Projection *proj, enum LatSgn ns)
{
  /*
   * Set proj to polar stereographic projection with
   * north or south pole indicated by lat of rpt and vertical meridian
   * given by lon of rpt.
   */

  /* Assign new values */
  sprintf(proj->descriptor, "\"Polar Stereographic\" %s hemisphere, ",
    ns == North ? "Northern" : "Southern");
  proj->prm.ns = ns;
  proj->projToLatLon = polarStereographicToLatLon;
  proj->latLonToProj = latLonToPolarStereographic;
  proj->latLonArrayToProj = latLonArrayToPolarSter;
}

GeoPoint
latLonToPolarStereographic(GeoPoint latLon, struct Projection * proj)
{
  GeoPoint projPoint;  
  double r, dLon;
  double nsFac = proj->prm.ns == North ? -1.0 : 1.0;
  
  /* Avoid opposite pole */
  latLon.ord =
    (proj->prm.ns == North && LatCmp(latLon.ord, -80.0) == South)
     ? -80.0 : (proj->prm.ns == South && LatCmp(latLon.ord, 80.0) == North)
                ? 80.0 : latLon.ord;

  r = 2.0 * REARTH * tan(M_PI_4 + latLon.ord * nsFac * 0.5 * RADPERDEG);
  dLon = latLon.abs * RADPERDEG;
  projPoint.abs =  r * sin(dLon);
  projPoint.ord = r * cos(dLon) * nsFac;
  return projPoint;
}

GeoPoint
polarStereographicToLatLon(GeoPoint projPt, struct Projection * proj)
{
  GeoPoint latLon;
  double rho, c;
  double nsFac = proj->prm.ns == North ? 1.0 : -1.0;

  rho = sqrt(projPt.abs * projPt.abs + projPt.ord * projPt.ord);
  if (rho == 0.0) {
    latLon.ord = 90.0 * nsFac;
    latLon.abs = 0.0;
    return latLon;
  }

  c = 2.0 * atan2(rho, 2.0 * REARTH);

  latLon.ord = asin(cos(c) * nsFac) * DEGPERRAD;
  latLon.abs = atan2(projPt.abs, projPt.ord * -nsFac) * DEGPERRAD;
  return latLon;
}

static GeoLineArray
latLonArrayToPolarSter(GeoLineArray latLonArray, struct Projection * proj)
{
  GeoLineArray projLineArray = NULL;	/* Return value */
  int nl, nls, np;			/* Loop parameters */
  GeoPoint latLon;
  GeoLine projLine = NULL;

  nls = GeoLineArrayNLines(latLonArray);
  if (   !(projLineArray = NewGeoLineArray(nls))
      || !(projLine = NewGeoLine(100)))
    goto error;

  for (nl = 0; nl < nls; nl++) {
    /* Loop through the lines.  Convert each line to a projection coordinates */
    GeoLine latLons = GetGeoLine(latLonArray, nl);

    for (np = 0; np < GeoLineNPts(latLons); np++) {
      latLon = GetGeoPoint(latLons, np);
      AppendGeoPoint(LatLonToProj(latLon, proj), projLine);
    }
    if (GeoLineNPts(projLine) > 1)
      AppendGeoLine(projLine, projLineArray);
    ClearGeoLine(projLine);
  }

  FreeGeoLine(projLine);
  return projLineArray;

  error:
  FreeGeoLine(projLine);
  FreeGeoLineArray(projLineArray);
  return NULL;
}

/************* Sinusoidal projection *************/

void
SetSinusoidalProj(struct Projection *proj, double rlon)
{
  /* Set proj to sinusoidal projection referenced to refLon */

  proj->prm.rlon = GwchLon(rlon);
  proj->projToLatLon = sinusoidalToLatLon;
  proj->latLonToProj = latLonToSinusoidal;
  proj->latLonArrayToProj = latLonArrayToCylProj;
  sprintf(proj->descriptor, "Sinusoidal centered at %.3f", proj->prm.rlon);
}

GeoPoint
latLonToSinusoidal(GeoPoint latLon, struct Projection * proj)
{
  double dlon;			/* Distance from reference point (deg lon) */
  GeoPoint projPoint;
  
  /* Determine offset from reference longitude in [-180.0, 180.0] */
  dlon = DomainLon(latLon.abs, proj->prm.rlon) - proj->prm.rlon;

  projPoint.abs = CMPERDEG * dlon * cos(RADPERDEG * latLon.ord);
  projPoint.ord = CMPERDEG * latLon.ord;
  return projPoint;
}

GeoPoint
sinusoidalToLatLon(GeoPoint projPoint, struct Projection * proj)
{
  GeoPoint latLon;

  latLon.ord = projPoint.ord * DEGPERCM;
  latLon.abs = DomainLon(projPoint.abs * DEGPERCM / cos(RADPERDEG * latLon.ord)
                           + proj->prm.rlon, proj->prm.rlon);
  return latLon;
}

/************* Orthographic projection *************/

void
SetOrthographic(struct Projection *proj, GeoPoint rpt)
{
  /* Set proj to orthographic projection with reference point rpt */
  proj->prm.refPt.pt = GwchLonPt(rpt);
  proj->prm.refPt.cosRLat = cos(rpt.ord * RADPERDEG);
  proj->prm.refPt.sinRLat = sin(rpt.ord * RADPERDEG);
  proj->latLonToProj = latLonToOrthographic;
  proj->projToLatLon = orthographicToLatLon;
  proj->latLonArrayToProj = latLonArrayToHemisphere;
  sprintf(proj->descriptor, "Orthographic centered at (%.3f %.3f)",
            proj->prm.refPt.pt.ord, proj->prm.refPt.pt.abs);
}

GeoPoint
latLonToOrthographic(GeoPoint latLon, struct Projection * proj)
{
  GeoPoint
    projPt,
    rpt = proj->prm.refPt.pt,
    nowhere = NOWHERE;
  double
    lat = latLon.ord * RADPERDEG,
    cosRLat = proj->prm.refPt.cosRLat,
    sinRLat = proj->prm.refPt.sinRLat,
    dlon = (latLon.abs - proj->prm.refPt.pt.abs) * RADPERDEG,
    dist;
  double
    coslat = cos(lat);

  dist = GeoDistance(rpt, latLon);
  if (DegCmp(dist, 90.0) == 1)
    return nowhere;
  projPt.abs = REARTH * coslat * sin(dlon);
  projPt.ord = REARTH * (cosRLat * sin(lat) - sinRLat * coslat * cos(dlon));
  return projPt;
}

GeoPoint
orthographicToLatLon(GeoPoint projPoint, struct Projection * proj)
{
  GeoPoint latLon;
  GeoPoint nowhere = NOWHERE;
  double
    r,		/* Polar radius */
    c;		/* Angular distance from latLon to ref point */
  double	/* Computational constants */
    cosC, sinC, ord,
    cosRLat = proj->prm.refPt.cosRLat,
    sinRLat = proj->prm.refPt.sinRLat;

  r = sqrt(projPoint.abs * projPoint.abs + projPoint.ord * projPoint.ord);
  if (r == 0.0) {
    return proj->prm.refPt.pt;
  }
  if (r / REARTH > 1.0) {
    return nowhere;
  }
  c = asin(r / REARTH);
  cosC = cos(c);
  sinC = sin(c);
  ord = cosC * sinRLat + (projPoint.ord * sinC * cosRLat / r);
  if (ord > 1.0) 
    return nowhere;
  latLon.ord = asin(ord) * DEGPERRAD;
  latLon.abs = proj->prm.refPt.pt.abs + atan2(projPoint.abs * sinC,
	                 (r * cosRLat * cosC - projPoint.ord * sinRLat * sinC))
			 * DEGPERRAD;
  latLon.abs = DomainLon(latLon.abs, proj->prm.refPt.pt.abs);
  return latLon;
}

/************* Functions used by more than one projection *************/

GeoLineArray
latLonArrayToCylProj(GeoLineArray latLonArray, struct Projection * proj)
{
  /*
   * Convert an array of latitude longitude lines to projection coordinates
   * where the projection is onto a cylinder that wraps around the Earth
   * with a discontinuity in longitude 180 deg from the reference longitude.
   * Lines are not allowed to cross the discontinuity.  They are broken in
   * two.  Returns a new geoLineArray with latLons in projection coordinates.
   * Storage for this array is dynamically allocated and must be freed by the
   * user.  Returns NULL on failure.
   */

  GeoLineArray projLineArray = NULL;	/* Return value */
  int nl, nls, np;			/* Loop parameters */
  double
    rlon = proj->prm.rlon,		/* Reference longitude */
    west = rlon - 180.0,
    east = rlon + 180.0,		/* West and east ends of domain */
    lon0, lon1;				/* Longitudes of last and current pt */
  GeoPoint
    latLon0, latLon1,			/* Points in the current segment */
    cross;				/* Point on domain boundary */
  GeoLine projLine = NULL;		/* Line being made right now */

  nls = GeoLineArrayNLines(latLonArray);
  if (   !(projLineArray = NewGeoLineArray(nls))
      || !(projLine = NewGeoLine(100)))
    goto error;

  for (nl = 0; nl < nls; nl++) {
    /* Loop through the lines.  Convert each line to a projection coordinates */
    GeoLine latLons = GetGeoLine(latLonArray, nl);

    latLon0 = GetGeoPoint(latLons, 0);
    AppendGeoPoint(LatLonToProj(latLon0, proj), projLine);
    for (np = 1; np < GeoLineNPts(latLons); np++) {
      latLon1 = GetGeoPoint(latLons, np);

      /* Do bookkeeping for segments crossing or touching boundaries */
      if (rlon > 0) {
        
	/* If rlon > 0, worry about segments crossing the western boundary */

	lon0 = DomainLon(latLon0.abs, west);
	lon1 = DomainLon(latLon1.abs, west);

	if (LonBtwn(west, lon0, lon1)) {

  	  /* Segment crosses longitude discontinuity */
  	  cross.ord = latLon0.ord + (latLon1.ord - latLon0.ord)
  	                            / (lon1 - lon0) * (west - lon0);
  	  if (LonCmp(lon0, west) == West) {
  	    /* Segment is crossing singularity from west to east */
  	    cross.abs = east;
            AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            if (GeoLineNPts(projLine) > 1)
              AppendGeoLine(projLine, projLineArray);
  	    ClearGeoLine(projLine);
  	    cross.abs = west;
            AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);
  	  } else {
  	    /* Segment is crossing singularity from east to west */
  	    cross.abs = west;
            AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            if (GeoLineNPts(projLine) > 1)
              AppendGeoLine(projLine, projLineArray);
  	    ClearGeoLine(projLine);
  	    cross.abs = east;
            AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);
  	  }

        } else if (LonCmp(lon0, west) == West && LonCmp(lon1, west) == PrMd) {

  	  /* Segment is terminating at eastern boundary */
	  cross.ord = latLon1.ord;
  	  cross.abs = east;
  	  AppendGeoPoint(LatLonToProj(cross, proj), projLine);
          if (GeoLineNPts(projLine) > 1)
            AppendGeoLine(projLine, projLineArray);
          ClearGeoLine(projLine);
  	  AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        } else if (LonCmp(lon0, west) == PrMd && LonCmp(lon1, west) == West) {

          /* Segment is starting at eastern boundary and going west */
          if (GeoLineNPts(projLine) > 1)
            AppendGeoLine(projLine, projLineArray);
          ClearGeoLine(projLine);
          latLon0.abs = east;
          AppendGeoPoint(LatLonToProj(latLon0, proj), projLine);
          AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        } else {

	  if (DegCmp(lon0, east) == 0 && DegCmp(lon1, west) == 0)
	    latLon1.abs = east;
	  if (DegCmp(lon0, west) == 0 && DegCmp(lon1, east) == 0)
	    latLon1.abs = west;
          AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        }

      } else {	/* rlon <= 0 */

	/* If rlon < 0, worry about segments crossing the eastern boundary */

  	lon0 = DomainLon(latLon0.abs, east);
  	lon1 = DomainLon(latLon1.abs, east);

	if (LonBtwn(east, lon0, lon1)) {

  	  /* Segment crosses longitude discontinuity */
  	  cross.ord = latLon0.ord + (latLon1.ord - latLon0.ord)
  	                            / (lon1 - lon0) * (east - lon0);
	  if (LonCmp(lon0, east) == West) {
	    /* Segment is crossing singularity from west to east */
	    cross.abs = east;
	    AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            if (GeoLineNPts(projLine) > 1)
              AppendGeoLine(projLine, projLineArray);
	    ClearGeoLine(projLine);
	    cross.abs = west;
	    AppendGeoPoint(LatLonToProj(cross, proj), projLine);
	    AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);
          } else {
	    /* Segment is crossing singularity from east to west */
	    cross.abs = west;
	    AppendGeoPoint(LatLonToProj(cross, proj), projLine);
            if (GeoLineNPts(projLine) > 1)
              AppendGeoLine(projLine, projLineArray);
	    ClearGeoLine(projLine);
	    cross.abs = east;
	    AppendGeoPoint(LatLonToProj(cross, proj), projLine);
	    AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);
          }

        } else if (LonCmp(lon0, east) == PrMd && LonCmp(lon1, east) == East) {

	  /* Segment is starting at western boundary and going east */
	  if (GeoLineNPts(projLine) > 1)
	    AppendGeoLine(projLine, projLineArray);
	  ClearGeoLine(projLine);
	  latLon0.abs = west;
	  AppendGeoPoint(LatLonToProj(latLon0, proj), projLine);
	  AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        } else if (LonCmp(lon0, east) == East && LonCmp(lon1, east) == PrMd) {

	  /* Segment is terminating at western boundary */
	  cross.ord = latLon1.ord;
  	  cross.abs = west;
  	  AppendGeoPoint(LatLonToProj(cross, proj), projLine);
          if (GeoLineNPts(projLine) > 1)
            AppendGeoLine(projLine, projLineArray);
          ClearGeoLine(projLine);
  	  AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        } else {

	  if (DegCmp(lon0, east) == 0 && DegCmp(lon1, west) == 0)
	    latLon1.abs = east;
	  if (DegCmp(lon0, west) == 0 && DegCmp(lon1, east) == 0)
	    latLon1.abs = west;
          AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

        }
      }
      latLon0 = latLon1;
    }
    if (GeoLineNPts(projLine) > 1)
      AppendGeoLine(projLine, projLineArray);
    ClearGeoLine(projLine);
  }

  FreeGeoLine(projLine);
  return projLineArray;

  error:
  FreeGeoLine(projLine);
  FreeGeoLineArray(projLineArray);
  return NULL;
}

GeoLineArray
latLonArrayToHemisphere(GeoLineArray latLonArray, struct Projection * proj)
{
  /*
   * Convert an array of latitude longitude lines to projection coordinates
   * where the projection is defined for the hemisphere containing the 
   * projection's reference point.  Segments do not cross the great circle
   * 90 degrees from the reference point.  E.g. orthographic.
   * Returns a new geoLineArray with latLons in projection coordinates.
   * Storage for this array is dynamically allocated and must be freed by the
   * user.  Returns NULL on failure.
   */

  GeoLineArray projLineArray = NULL;	/* Return value */
  int nl, nls, np;			/* Loop parameters */
  GeoPoint
    rpt = proj->prm.refPt.pt,		/* Reference point */
    latLon0, latLon1;			/* Points in the current segment */
  int in0, in1;				/* True if point in domain */
  GeoLine projLine = NULL;		/* Line being made right now */

  nls = GeoLineArrayNLines(latLonArray);
  if (   !(projLineArray = NewGeoLineArray(nls))
      || !(projLine = NewGeoLine(100)))
    goto error;

  for (nl = 0; nl < nls; nl++) {
    /* Loop through the lines.  Convert each line to a projection coordinates */
    GeoLine latLons = GetGeoLine(latLonArray, nl);

    latLon0 = GetGeoPoint(latLons, 0);

    /* latLon0 is in domain if distance from refPt < 90 degrees */
    in0 = DegCmp(GeoDistance(rpt, latLon0), 90.0) == -1;

    if (in0) {
      AppendGeoPoint(LatLonToProj(latLon0, proj), projLine);
    }

    for (np = 1; np < GeoLineNPts(latLons); np++) {
      latLon1 = GetGeoPoint(latLons, np);

      /* latLon1 is in domain if distance from refPt < 90 degrees */
      in1 = DegCmp(GeoDistance(rpt, latLon1), 90.0) == -1;

      if (in0 && in1) {

	/* Segment is in projection domain */
        AppendGeoPoint(LatLonToProj(latLon1, proj), projLine);

      } else if ((in0 && !in1) || (!in0 && in1)) {

        /* Segment straddles projection domain.  Find boundary point. */
	CartPoint
	  pt0 = LatLonToCart(latLon0),	/* latLon0 in Cartesian coord's */
	  pt1 = LatLonToCart(latLon1),	/* latLon1 in Cartesian coord's */
	  mid,				/* Point half way between pt0 and pt1 */
	  rptc = LatLonToCart(rpt),	/* refpoint in Cartesian coord's */
	  pt0Xpt1,			/* Cross product of pt0 and pt1 */
	  bndPtC;			/* Boundary point Cartesian coord's */
	GeoPoint bndPt;			/* Boundary point as lat-lon */
        double
	  nmlz,				/* Normalizing factor */
	  dp, dm;	/* Cartesian distance between mid point and +-bndPt */

	/* Cross product of segment end points */
        pt0Xpt1.x = pt0.y * pt1.z - pt0.z * pt1.y;
        pt0Xpt1.y = pt0.z * pt1.x - pt0.x * pt1.z;
        pt0Xpt1.z = pt0.x * pt1.y - pt0.y * pt1.x;

	/* Cross product of pt0Xpt1 and rptc points toward (and away from)
	 * bound point */
        bndPtC.x = pt0Xpt1.y * rptc.z - pt0Xpt1.z * rptc.y;
        bndPtC.y = pt0Xpt1.z * rptc.x - pt0Xpt1.x * rptc.z;
        bndPtC.z = pt0Xpt1.x * rptc.y - pt0Xpt1.y * rptc.x;

	nmlz = 1.0/sqrt(bndPtC.x*bndPtC.x+bndPtC.y*bndPtC.y+bndPtC.z*bndPtC.z);
        bndPtC.x *= nmlz;
        bndPtC.y *= nmlz;
        bndPtC.z *= nmlz;

	if (bndPtC.x == 0.0 && bndPtC.y == 0.0) {
	  /* Boundary point is at a pole */
          bndPt.ord = (bndPtC.z > 0) ? 90.0 : -90.0;
          bndPt.abs = 0.0;		/* Arbitrary lon at pole */
	}

	/* Boundary point is either bndPt or the point opposite it
	 * Take the one close to mean of pt0 and pt1 */
        mid.x = 0.5 * (pt0.x + pt1.x);
        mid.y = 0.5 * (pt0.y + pt1.y);
        mid.z = 0.5 * (pt0.z + pt1.z);
	dp = (mid.x - bndPtC.x) * (mid.x - bndPtC.x)
	     + (mid.y - bndPtC.y) * (mid.y - bndPtC.y)
	     + (mid.z - bndPtC.z) * (mid.z - bndPtC.z);
	dm = (mid.x + bndPtC.x) * (mid.x + bndPtC.x)
	     + (mid.y + bndPtC.y) * (mid.y + bndPtC.y)
	     + (mid.z + bndPtC.z) * (mid.z + bndPtC.z);
        if (dm < dp) {
	  bndPtC.x = -bndPtC.x;
	  bndPtC.y = -bndPtC.y;
	  bndPtC.z = -bndPtC.z;
	}

	/* Get boundary point as lat-lon */
        bndPt.ord = atan(bndPtC.z / sqrt(bndPtC.x*bndPtC.x + bndPtC.y*bndPtC.y))
	            * DEGPERRAD;
        bndPt.abs = atan2(bndPtC.y, bndPtC.x) * DEGPERRAD;

	if (in0 && !in1) {
	  /* Segment is leaving projection domain.  Close out the line */
          AppendGeoPoint(LatLonToProj(bndPt, proj), projLine);
          if (GeoLineNPts(projLine) > 1)
            AppendGeoLine(projLine, projLineArray);
          ClearGeoLine(projLine);
        } else {
	  /* Segment is entering projection domain.  Start a new line. */
          AppendGeoPoint(LatLonToProj(bndPt, proj), projLine);
        }

      }  /* endif segment straddles domain boundary */

      latLon0 = latLon1;
      in0 = in1;

    }  /* for np */

    if (GeoLineNPts(projLine) > 1)
      AppendGeoLine(projLine, projLineArray);
    ClearGeoLine(projLine);

  }  /* for nl */

  FreeGeoLine(projLine);
  return projLineArray;

  error:
  FreeGeoLine(projLine);
  FreeGeoLineArray(projLineArray);
  return NULL;
}
