/*
 * getrgn - Pipe points within chosen region from file (or stdin) to stdout.
 *
 * Copyright (C) 1998 Gordon D. Carrie
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Please address questions and bug reports to gcarri@mail.txcyber.com
 *
 **********************************************************************
 *
 * Usage:
 *   getrng [-n north] [-e east] [-s south] [-w west] [-f file]
 * Input and output are xdr streams in the following format:
 * First record: descriptor string
 * Then for each line in the outline data set
 * 	Int record giving number of points in the line
 * 	For each geographical point on the line
 *	  Float record giving point's latitude
 *	  Float record giving point's longitude
 * If no file is indicated, standard input is used.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <rpc/rpc.h>

char *progName;

typedef struct {
  float lat, lon;
} GeoPoint;

/* Array of GeoPoints */
typedef struct
{
  int      npts;		/* Number of points actually in array */
  int      maxPts;		/* Number of points allowed in allocation */
  GeoPoint *points;		/* Array of points */
} PtArray;

/* "Rectangular" region */
typedef struct
{
  float north;
  float east;
  float south;
  float west;
} Bounds;

typedef enum Region {inside, outside} Region;

/* Global function declarations */

void
ReadPoint(XDR *xdrIn, GeoPoint *point);

void
WritePoints(XDR *xdrOut, PtArray *outArray);

Region
GetRgn(GeoPoint geoPoint, Bounds bounds);

void
AppendPoint(GeoPoint geoPoint, PtArray *ptArray);

GeoPoint
BoundPt(GeoPoint geoPt1, GeoPoint geoPt2, Bounds bounds);

int
LatIntcpt(GeoPoint geoPt1, GeoPoint geoPt2, float lat, Bounds bounds,
	  GeoPoint *geoPoint);

int
LonIntcpt(GeoPoint geoPt1, GeoPoint geoPt2, float lon, Bounds bounds,
	  GeoPoint *geoPoint);
void
FailOut(char *message);

  /******************** Start of function definitions  ********************/

int
main(int argc, char *argv[])
{
  char
    *descriptor = NULL,		/* Human readable first record of file	*/
    **arg;			/* Loop pointer			        */

  Bounds bounds = {90.0, 180.0, -90.0, -180.0};

  XDR xdrIn, xdrOut;		/* Input and output xdr streams		*/

  int
    npts,			/* Number of points in a line	        */
    n;				/* Loop index                           */

  GeoPoint
    prevPt,			/* Previous point read in               */
    curPt;			/* Current point                        */

  PtArray outArray = {0, 0, NULL};
  
  Region
    prevRgn,			/* Region containing prevPt             */
    curRgn;			/* Region containing curPt              */

  /**** End of declarations ****/

  progName = argv[0];

  /* Parse command line */
  for (arg = argv + 1; *arg; arg++)
    {
      if (strcmp(*arg, "-n") == 0)
	{
	  arg++;
	  if (sscanf(*arg, "%f", &bounds.north) != 1)
	    {
	      fprintf(stderr, "%s: Could not read %s\n", progName, *arg);
	      exit(1);
	    }
	}
      else if (strcmp(*arg, "-e") == 0)
	{
	  arg++;
	  if (sscanf(*arg, "%f", &bounds.east) != 1)
	    {
	      fprintf(stderr, "%s: Could not read %s\n", progName, *arg);
	      exit(1);
	    }
	}
      else if (strcmp(*arg, "-s") == 0)
	{
	  arg++;
	  if (sscanf(*arg, "%f", &bounds.south) != 1)
	    {
	      fprintf(stderr, "%s: Could not read %s\n", progName, *arg);
	      exit(1);
	    }
	}
      else if (strcmp(*arg, "-w") == 0)
	{
	  arg++;
	  if (sscanf(*arg, "%f", &bounds.west) != 1)
	    {
	      fprintf(stderr, "%s: Could not read %s\n", progName, *arg);
	      exit(1);
	    }
	}
      else if (strcmp(*arg, "-f") == 0)
	{
	  arg++;
	  /* Associate stdin with file */
	  if (!freopen(*arg, "r", stdin))
	    {
	      fprintf(stderr, "%s: Could not open %s\n", progName, *arg);
	      exit(1);
	    }
	}
      else
	{
	  fprintf(stderr, "%s: Unknown option: %s\n", progName, *arg);
	  exit(1);
	}
    }

  if (bounds.west >= bounds.east || bounds.south >+ bounds.north)
    {
      fprintf(stderr, "%s: Bad bounds: w=%8.3f e=%8.3f s=%8.3f n=%8.3f\n",
	      progName, bounds.west, bounds.east, bounds.south, bounds.north);
      exit(1);
    }

  /* Identify standard input and output as xdr streams */
  xdrstdio_create(&xdrIn, stdin, XDR_DECODE);
  xdrstdio_create(&xdrOut, stdout, XDR_ENCODE);
	
  /* Read and write descriptor */
  if (xdr_string(&xdrIn, &descriptor, INT_MAX) == FALSE)
    FailOut("Descriptor read failed.");
  if (xdr_string(&xdrOut, &descriptor, INT_MAX) == FALSE)
    FailOut("Descriptor write failed.");
  free(descriptor);

  /* Read and send lines until no more input */
  while (xdr_int(&xdrIn, &npts))
    {
      ReadPoint(&xdrIn, &prevPt);
      if ((prevRgn = GetRgn(prevPt, bounds)) == inside)
	AppendPoint(prevPt, &outArray);
      for (n = 1; n < npts; n++)
	{
	  ReadPoint(&xdrIn, &curPt);
	  curRgn = GetRgn(curPt, bounds);
	  if (prevRgn == inside && curRgn == inside)
	    {
	      /* Moving around inside region.  Add point to output array */
	      AppendPoint(curPt, &outArray);
	    }
	  else if (prevRgn == inside && curRgn == outside)
	    {
	      /* Moved outside region.  Append boundary point, send and */
	      /* reset output array */
	      AppendPoint(BoundPt(prevPt, curPt, bounds), &outArray);
	      WritePoints(&xdrOut, &outArray);
	    }
	  else if (prevRgn == outside && curRgn == inside)
	    {
	      /* Moved inside region.  Start output array with boundary */
	      /* point and current point */
	      AppendPoint(BoundPt(prevPt, curPt, bounds), &outArray);
	      AppendPoint(curPt, &outArray);
	    }
	  prevPt = curPt;
	  prevRgn = curRgn;
	}
      if (outArray.npts)	/* If outArray has points, send them */
	WritePoints(&xdrOut, &outArray);
    }
  xdr_destroy(&xdrOut);
  return 0;
}

void
ReadPoint(XDR *xdrIn, GeoPoint *point)
{
  /* Read one GeoPoint from xdrIn.  Exit on failure */
  if ( !(xdr_float(xdrIn, &point->lat) && xdr_float(xdrIn, &point->lon)))
    FailOut("Read point failed.");
}

void
WritePoints(XDR *xdrOut, PtArray *outArray)
{
  /* Send an array of GeoPoints to xdrOut */
  
  int n;
  
  if ( !xdr_int(xdrOut, &outArray->npts))
    FailOut("Could not write points.");
  for (n = 0; n < outArray->npts; n++)
    if ( !(   xdr_float(xdrOut, &outArray->points[n].lat)
	   && xdr_float(xdrOut, &outArray->points[n].lon)))
      FailOut("Could not write points.");
  outArray->npts = 0;
}

Region
GetRgn(GeoPoint geoPoint, Bounds bounds)
{
  /* Indicate whether geoPoint is inside or outside bounds */
  
  return (geoPoint.lat > bounds.north || geoPoint.lat < bounds.south
	  || geoPoint.lon < bounds.west || geoPoint.lon > bounds.east)
    ? outside : inside;
}

void
AppendPoint(GeoPoint geoPoint, PtArray *ptArray)
{
  /* Add geoPoint to ptArray */

  int incr = 10;		/* Amount to increase points allocation */

  /* Allocate or reallocate memory */
  if ( !ptArray->points)
    {
      if ( !(ptArray->points = (GeoPoint *)malloc(incr * sizeof(PtArray))))
	  FailOut("Memory allocation failed.");
      ptArray->maxPts = incr;
      ptArray->npts = 0;
    }
  else if (ptArray->npts + 1 > ptArray->maxPts
	   && !(ptArray->points
		= (GeoPoint *)realloc(ptArray->points,
				      (ptArray->maxPts += incr)
				      * sizeof(GeoPoint))))
    {
      FailOut("Memory allocation failed.");
    }
  ptArray->points[ptArray->npts++] = geoPoint;
}

GeoPoint
BoundPt(GeoPoint geoPt1, GeoPoint geoPt2, Bounds bounds)
{
  /* Return point on bounds between geoPt1 and geoPt2 */

  GeoPoint geoPoint;		/* Hold return value */

       if ((geoPt1.lat - bounds.north) * (bounds.north - geoPt2.lat) >= 0
	   && LatIntcpt(geoPt1, geoPt2, bounds.north, bounds, &geoPoint))
    return geoPoint;
  else if ((geoPt1.lat - bounds.south) * (bounds.south - geoPt2.lat) >= 0
	   && LatIntcpt(geoPt1, geoPt2, bounds.south, bounds, &geoPoint))
    return geoPoint;
  else if ((geoPt1.lon - bounds.west) * (bounds.west - geoPt2.lon) >= 0
	   && LonIntcpt(geoPt1, geoPt2, bounds.west, bounds, &geoPoint))
    return geoPoint;
  else if ((geoPt1.lon - bounds.east) * (bounds.east - geoPt2.lon) >= 0
	   && LonIntcpt(geoPt1, geoPt2, bounds.east, bounds, &geoPoint))
    return geoPoint;

  return geoPoint;
}

int
LatIntcpt(GeoPoint geoPt1, GeoPoint geoPt2, float lat, Bounds bounds,
	  GeoPoint *geoPoint)
{
  /* Find intersection of line through geoPt1 and geoPt2 on latitude line */
  /* at lat between bounds.west and bounds.east.  If intersection exists, */
  /* puts it into geoPoint and returns true, o/w returns false            */

  float lon;

  if (geoPt2.lat == geoPt1.lat)
    return 0;

  lon = geoPt1.lon + (geoPt2.lon - geoPt1.lon) / (geoPt2.lat - geoPt1.lat)
     * (lat - geoPt1.lat);

  if (bounds.west < lon && lon < bounds.east)
    {
      geoPoint->lat = lat;
      geoPoint->lon = lon;
      return 1;
    }

  return 0;
}

int
LonIntcpt(GeoPoint geoPt1, GeoPoint geoPt2, float lon, Bounds bounds,
	  GeoPoint *geoPoint)
{
  /* Find intersection of line through geoPt1 and geoPt2 on meridian        */
  /* at lon between bounds.south and bounds.north.  If intersection exists, */
  /* puts it into geoPoint and returns true, o/w returns false              */

  float lat;

  if (geoPt2.lon == geoPt1.lon)
    return 0;

  lat = geoPt1.lat + (geoPt2.lat - geoPt1.lat) / (geoPt2.lon - geoPt1.lon)
    * (lon - geoPt1.lon);

  if (bounds.south < lat && lat < bounds.north)
    {
      geoPoint->lat = lat;
      geoPoint->lon = lon;
      return 1;
    }

  return 0;
}

void
FailOut(char *message)
{
  fprintf(stderr, "%s: %s\n", progName, message);
  exit(1);
}
